% =========================================================================
% -------------------------------------------------------------------------
%
%        Estimation of Distribution using Random Projection Technique
%
% -------------------------------------------------------------------------
% =========================================================================
%    Version : 1.0 (7 March 2013)
%    Since   : 27 June 2012
%    Authors : Ata Kaban, Jakramate Bootkrajang, Robert Durrant. 
%              School of Computer Science, The University of Birmingham
%    Contact : J.Bootkrajang@cs.bham.ac.uk
% =========================================================================
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with this program.  If not, see <http://www.gnu.org/licenses/>.
% =========================================================================
% INPUT:
%   funOjb : Objective function handle  (i.e. @your_function_name)
%   options: A structure of options, possible fields are
%      REQUIRED
%         options.ub
%             Upperbound on possible solution to the funObj
%         options.lb
%             Lowerbound on possible solution to the funObj
%         options.optmType
%             'min' = to minimise the funObj
%             'max' = to maximise the funObj
%      OPTIONAL
%         options.verbose
%             Print useful information during the run
%             [default = false]
%         options.maxFE
%             Maximum function evaluation allowed
%             [default = 100000]
%         options.topSize
%             The size of best individuals retained
%             [default = 0.25 * size(x,1)]
%         options.k
%             Projection dimensionality 
%             [default = 3]
%         options.rpmSize
%             Number of projections per one individual
%             [default = 1000]
%     x:  initial population in the form of an 'n x d' matrix where 'n' is the
%         number of populations and 'd' is the dimensionality of the problem
%         (note: variable 'popSize' will be set to 'n');
%     varargin: Arguments to 'funObj'
%    DEFAULT PARAMETERS
%          Random matrices are generated from ~ N(0,sqrt(1/d)) with FULL
%          covariance matrix.
%--------------------------------------------------------------------------
% OUTPUT:
%     optm_x : Optimal solution to the objective
%     fv     : Optimal objective value;
%     stats  : Info. about dynamics of evolution
%         stats.fv  = best fitness in each generation
%         stats.dev = std of fitnesses in the pop in each generation
%         stats.gap = distance of the best individual from x*
%     options: Options used (for follow up analysis)
% =========================================================================


function [optm_x fv stats options] = rp_eda(funObj, options, x, varargin)

%% preprocessing stage
[popSize d]     = size(x);
options.popSize = popSize;

if ~isfield(options, 'ub');
    error('Please provide upperbound on function');
end

if ~isfield(options, 'lb');
    error('Please provide lowerbound on function');
end

if ~isfield(options, 'optmType')
    error('Please provide optimisation type {min, max}');
end

% print some useful info?
if isfield(options, 'verbose')
    verbose = options.verbose;
else
    verbose = false;
    options.verbose = verbose;
end

% define maximum generation
if isfield(options, 'maxFE')
    maxFE = options.maxFE;
else
    maxFE = 100000;
    options.maxFE = maxFE;
end

% define best population size
if isfield(options, 'topSize')
    topSize = options.topSize;
    if topSize > popSize
        disp('Warning: topSize is larger than popSize, will use, popSize/10');
        topSize = 0.25 * popSize;
        options.topSize = topSize;
    end
else
    topSize = 0.25 * popSize;
    options.topSize = topSize;
end

% define projection dimensionality
if isfield(options, 'k')
    k = options.k;
    % fail safe check
    if k > d
        disp('Warning: projection dimension is larger than original dimension.');
    end
else
    k = 3;
    options.k = k;
end

% define cardinality of RPM set;
if isfield(options, 'rpmSize')
    rpmSize = options.rpmSize;
else
    rpmSize = 1000;
    options.rpmSize = rpmSize;
end

% # of projections per 1 loop, large values seem to freeze MATLAB, 100~200
% is a reasonable choice
lstep = 500;
xtmp  = zeros(popSize,k,lstep);

% all initial pop are top_x
top_x  = x;
usedFE = 0;
iter   = 1;
stats.gap=zeros(1,10000); stats.dev=stats.gap;stats.fv=stats.gap;

% precompute some constants
sq_1_d = sqrt(1/d);
sq_d_k = sqrt(d/k);
sq_rpmSize = sqrt(rpmSize);

%% The optimisation begins

while true

    mu  = sum(top_x,1)/size(top_x,1);
    MU  = bsxfun(@minus, top_x, mu);
       
    xp  = zeros(popSize,d);
    
    for c=1:lstep:rpmSize
        % projecting down to k-dim, each proj contributes 'popSize' individuals
        R   = randn(d, k, lstep,'single') .* sq_1_d;
        pjx = mtimesx(MU,R);
        
        for t = 1:lstep
            cvr = cov(pjx(:,:,t));
            xtmp(:,:,t) = mvnrnd(zeros(1,k), cvr, popSize);
        end

        % projecting back to the original dataspace
        Xp = mtimesx(xtmp,R,'T');
        xp = xp + sum(Xp,3);
    end
         
    xp = xp/rpmSize;
    xp = bsxfun(@minus, xp, sum(xp,1)/popSize);
    xp = xp * sq_rpmSize * sq_d_k;
    xp = bsxfun(@plus, xp, mu);   
    xp(xp>options.ub) = options.ub;
    xp(xp<options.lb) = options.lb;    
    x = xp;
        
    % keeping the best individual
    x(1,:) = top_x(1,:);
        
    % measuring the fitness of the current population
    remainFE  = maxFE - usedFE;
    if remainFE == 0
        break;
    elseif  remainFE - popSize >= 0
        [fval xstar] = funObj(x, varargin{:});
        usedFE       = usedFE + popSize;
    else
        [fval xstar] = funObj(x(1:remainFE,:),varargin{:});
        usedFE       = usedFE + remainFE;
    end
    
    % selecting top 'topSize' individuals based on fitness values
    if strcmp(options.optmType, 'max')
        [void, b] = sort(fval, 'descend');  % maximising
    else
        [void, b] = sort(fval, 'ascend');   % minimising
    end
    top_x  = x(b(1:topSize),:);             % select top individuals
    optm_x = x(b(1),:);                     % save optimal parameter
    fv     = fval(b(1));                    % save optimal func. value        
     
    % collecting stats
    stats.fv(iter)  = fv;
    stats.dev(iter) = std(fval);
    stats.gap(iter) = norm(xstar - optm_x);  
        
    %%========================  VISUALISATION  ===========================        
    if verbose
        figure(1)
        % plotting function value
        subplot(2,1,1);
        plot(stats.fv(1:iter),'x','LineWidth', 2);
        ylim([stats.fv(iter)-200, stats.fv(iter)+200]);
        title(['Best FV:' num2str(fv) '   Tr(cov)/d:' num2str(trace(cov(x))/d)]);
        xlabel('generations')
        ylabel('Function value');
        
        % plotting distance of optm_x to xstar
        subplot(2,1,2);
        gap = norm(optm_x - xstar);
        plot(stats.gap(1:iter),'x', 'LineWidth', 2);
        title(['Current distance from bestX to xStar:' num2str(gap)]);
        xlabel('# of function evaluations');
        ylabel('Distance (L2-norm)');drawnow        
    end
    
    iter = iter + 1;
    
end % end while